// ****************** MUST INCLUDE LINES BELOW*****************************
#include <lib6310.h>
// Init library and set DAC resolution, ADC resolution, serial channel
#define MYSERIAL Serial
lib6310 my6310(12,12,&Serial);
// Set up the analog-to-digital library, with a buffersize
#define sampleBufN 500
LIB6310_ADC(sampleBufN+100);  // ADC sample buffer of 100+sampleBufN
// ****************** MUST INCLUDE LINES ABOVE*******1**********************

// Constants for Users to Modify 
// ************************************************************
#define AngleOffset 0.0 // Adjust if measured angle is not zero when horizontal. 
#define Psmooth 0.0 // Toggle smoothing 0: no smooth, 0.9->0.999: less->more smooth
#define DeltaT_Microsecs 1000 // Update period in microseconds
#define nomCommand 0.4 // Motor command that nominally hold arm level.
#define MbackMax 250  // Maximum value of mBack
#define NUMMOTORS 1 // Number of propeller motors
#define nomCmdDeltaR 0.0 // Amount to increase right motor to balance arm

// Indexes for values being sent from plotter send window.
// e.g. to set KP to 0.3, type "3 0.3" in the send window and click send, 
// e.g. to set the input period to two seconds, type "5 0.5" and click send.
#define EXTRABACK 0  // dx/dt is approx (x[n]-x[n-m])/(DeltaT*m), m = 1+extraBack
#define KD 1         // Derivative Gain
#define KI 2         // Integral Gain
#define KP 3         // Proportional Gain
#define SKIPS 4      // Number of sample points skipped between plotted points
#define FREQ 5       // Input frequency 
#define AMP 6        // Input peak amplitude, Amp>0:step inputs, Amp<0:sine inputs

// offsets for parameters sendable from plotter window.
#define offsetExtraBack 0
#define offsetKd 0.0
#define offsetKi 0.0
#define offsetKp 0.0
#define offsetNumSkips 9
#define offsetAmp 0.0
#define offsetFreq 0.0

// ************************************************************************
// End of Constants and code for Users to Modify 
// SEE END OF FILE FOR USER-ALTERABLE CONTROL LOOP!!

// *******SETTING UP THE MONITOR DISPLAY (SLIDERS AND PLOTS). *************
// Sliders or settings from plotter window.
String sliders = "S~ExtraBack~0~20~S~Kd~0~5~S~Ki~0~100~S~Kp~0~10~S~Skips~0~50~S~Freq~0.1~20~S~Amp~-3~3~";

// Plots: USE SPACES BETWEEN LABELS FOR MULTIPLOTS!!
String measDesPlot = "P~Theta_m Theta_d Cmd~-1~1~500~3~"; 
String restPlot = "P~DEdt/10 ErrSum~-1~1~500~2~";


#define PLOT_TM 0
#define PLOT_TD 1
#define PLOT_CMD 2
#define PLOT_DDT 3
#define PLOT_SUM 4

String monPlots = measDesPlot + restPlot;

// Configuration string with surrounding limits.
String config_message = sliders + monPlots;

// Control loop setup
// ************************************************************************

// Motor PWM and Sensor defines.ot + cmdPlot
elapsedMicros loopTimer = 0;  // Loop runs at least 10xsecond
float deltaT = 0.0;
float headRoom;

// Get Theta in radians
float getTheta() {
  uint16_t numSamp = 9;
  float theta = 0.0;
  for(int i = numSamp; i > 0; i--) {
    theta += 2*PI*(my6310.adcGetAvg(IN4) - 0.5);
  }
  theta = theta/float(numSamp) - AngleOffset;
  return(theta);
}

// Approximates d(x)/dt as (x[n] - x[n-m])/(m*deltaT)
float dXdt(float x, int m)
{
  static float xBack[MbackMax+1];  // Stores past errors for deriv approx

  m = max(m,1); // Make sure m is at least 1.

  // save past errors.
  for (int i = MbackMax; i > 0; i--) xBack[i] = xBack[i-1];
  xBack[0] = x;

  // Calculate approximation to d(x)/dt
  float dxDtVal = (x- xBack[m])/(float(m)*deltaT); 

  return(dxDtVal);
}


float errSum = 0.0;  // Stores accumulated errors.
float lastKi = 0.0;  // Used to zero sum when Ki changes.

// Computes the deltaT weight sum of errors.
float integX(float x, float Ki)
{
  static float xSum = 0.0;
  static float lastKi = 0.0;

  if(Ki > 0) {
    if(lastKi != Ki) xSum = 0.0;
    float SumMax = 100/(0.1 + Ki);
    xSum += x * deltaT; 
    xSum = max(min(SumMax,xSum), -SumMax);
  }
  lastKi = Ki;
  return(xSum);
}


void setup() {
  my6310.setup(config_message);
  my6310.startPWM(0.98); // start parallel drive pwm, A+B in sync.
  my6310.setupADC(8,false);
  loopTimer = 0;
  deltaT = float(DeltaT_Microsecs)*1.0e-6;
  headRoom = DeltaT_Microsecs;
}

//********************** Main USER ALTERABLE Control Loop **********************
//**************************************************************************
int skipCounter = 0;  // Counts plot-skipped samples.

void loop() {  
  my6310.startloop(); // Initialize and check for usb inputs.

  // Get the nominal propeller command from the analog input.
  float nomCmd = nomCommand*(1.0 + 2.0*(my6310.analog2param(IN8) - 0.5));

  // Get send window parameters.
  int extraBack = offsetExtraBack + my6310.getSlider(EXTRABACK); 
  float Kp = offsetKp + my6310.getSlider(KP); 
  float Kd = offsetKd + my6310.getSlider(KD); 
  float Ki = offsetKi + my6310.getSlider(KI);
  int numSkips = offsetNumSkips + my6310.getSlider(SKIPS);
  float freq =  offsetFreq + my6310.getSlider(FREQ);
  float amp = offsetAmp + my6310.getSlider(AMP);

  // Headroom is faction of period left until next sensor read.
  //float headRoom = float(DeltaT_Microsecs - int(loopTimer));

  // Wait exactly one deltaT since last sensor read.
  while (int(loopTimer) < DeltaT_Microsecs) {};
  loopTimer = 0;

  // Get the desired and measured arm angle.
  float theta_d = my6310.updateDesired(amp,freq,Psmooth);
  float theta_a = getTheta();

  // Compute the error and its approximate derivative and integral.
  float Err = (theta_d - theta_a);
  float dErrDt = dXdt(Err, 1+extraBack);
  float intErr = integX(Err, Ki);

  // Compute the offset-from-nominal motor cmd 
  float dCmd = Kp*Err + Kd*dErrDt + Ki*intErr;

  // Bound command and write to H-bridge 
  dCmd = my6310.scaleAndBound(dCmd, 1.0-nomCmd, 1.0); // Ensure symmetric bound.
  float cmdClip = my6310.hbridgeBipolar(nomCmd + dCmd, B2L, B1L); // B1L <-> B2L swaps direction

#if (NUMMOTORS == 2) 
  my6310.hbridgeBipolar((nomCmd - dCmd) + nomCmdDeltaR, A2R, A1R); // A1R <-> A2R swaps direction
#endif

  // Display the results 
  if(skipCounter > 0) skipCounter--;
  else {
    skipCounter = numSkips;
    // Send data to monitor.
    float loopStatus[5];
    loopStatus[PLOT_TM] = theta_a;
    loopStatus[PLOT_TD] =  theta_d; //headRoom;
    loopStatus[PLOT_CMD] = cmdClip;
    loopStatus[PLOT_DDT] = dErrDt/10;
    loopStatus[PLOT_SUM] = intErr/(0.1 + Ki);

    my6310.sendStatus(5,loopStatus); 
  }
}




  


  
